#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional

from neptune.new.internal.containers.storage_queue import StorageQueue
from neptune.new.internal.operation import Operation
from neptune.new.internal.operation_processors.operation_processor import OperationProcessor


class OfflineOperationProcessor(OperationProcessor):

    def __init__(self, queue: StorageQueue[Operation]):
        self._queue = queue

    def enqueue_operation(self, op: Operation, wait: bool) -> None:
        # pylint: disable=unused-argument
        self._queue.put(op)

    def wait(self):
        self.flush()

    def flush(self):
        self._queue.flush()

    def start(self):
        pass

    def stop(self, seconds: Optional[float] = None):
        pass
